/*
 * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file fd.c Helper functions for file descriptor handling. */

#include <sys/types.h>
#include <dirent.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/select.h>

#include "log.h"
#include "osl.h"
#include "error.h"

/**
 * Write a buffer to a file descriptor, re-write on short writes.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to be sent.
 * \param len The length of \a buf.
 *
 * \return Standard. In any case, the number of bytes that have been written is
 * stored in \a len.
 */
int write_all(int fd, const char *buf, size_t *len)
{
	size_t total = *len;

	assert(total);
	*len = 0;
	while (*len < total) {
		int ret = write(fd, buf + *len, total - *len);
		if (ret == -1)
			return -ERRNO_TO_ERROR(errno);
		*len += ret;
	}
	return 1;
}

/**
 * Wrapper for the open(2) system call.
 *
 * \param path The filename.
 * \param flags The usual open(2) flags.
 * \param mode Specifies the permissions to use.
 *
 * The mode parameter must be specified when O_CREAT is in the flags, and is
 * ignored otherwise.
 *
 * \return The file descriptor on success, negative on errors.
 *
 * \sa open(2).
 */
int osl_open(const char *path, int flags, mode_t mode)
{
	int ret = open(path, flags, mode);

	if (ret >= 0)
		return ret;
	return -ERRNO_TO_ERROR(errno);
}

/**
 * Wrapper for chdir(2).
 *
 * \param path The specified directory.
 *
 * \return Standard.
 */
static int para_chdir(const char *path)
{
	int ret = chdir(path);

	if (ret >= 0)
		return 1;
	return -ERRNO_TO_ERROR(errno);
}

/**
 * Save the cwd and open a given directory.
 *
 * \param dirname Path to the directory to open.
 * \param dir Result pointer.
 * \param cwd File descriptor of the current working directory.
 *
 * \return Standard.
 *
 * Opening the current directory (".") and calling fchdir() to return is
 * usually faster and more reliable than saving cwd in some buffer and calling
 * chdir() afterwards.
 *
 * If \a cwd is not \p NULL "." is opened and the resulting file descriptor is
 * stored in \a cwd. If the function returns success, and \a cwd is not \p
 * NULL, the caller must close this file descriptor (probably after calling
 * fchdir(*cwd)).
 *
 * On errors, the function undos everything, so the caller needs neither close
 * any files, nor change back to the original working directory.
 *
 * \sa getcwd(3).
 *
 */
int para_opendir(const char *dirname, DIR **dir, int *cwd)
{
	int ret;

	if (cwd) {
		ret = osl_open(".", O_RDONLY, 0);
		if (ret < 0)
			return ret;
		*cwd = ret;
	}
	ret = para_chdir(dirname);
	if (ret < 0)
		goto close_cwd;
	*dir = opendir(".");
	if (*dir)
		return 1;
	ret = -ERRNO_TO_ERROR(errno);
/* Ignore return value of fchdir() and close(). We're busted anyway. */
	if (cwd)
		fchdir(*cwd);
close_cwd:
	if (cwd)
		close(*cwd);
	return ret;
}

/**
 * A wrapper for fchdir().
 *
 * \param fd An open file descriptor.
 *
 * \return Standard.
 */
int para_fchdir(int fd)
{
	if (fchdir(fd) < 0)
		return -ERRNO_TO_ERROR(errno);
	return 1;
}

/**
 * A wrapper for mkdir(2).
 *
 * \param path Name of the directory to create.
 * \param mode The permissions to use.
 *
 * \return Standard.
 */
int para_mkdir(const char *path, mode_t mode)
{
	if (!mkdir(path, mode))
		return 1;
	return -ERRNO_TO_ERROR(errno);
}

/**
 * Open a file and map it into memory.
 *
 * \param path Name of the regular file to map.
 * \param open_mode Either \p O_RDONLY or \p O_RDWR.
 * \param map On success, the mapping is returned here.
 * \param size size of the mapping.
 * \param fd_ptr The file descriptor of the mapping.
 *
 * If \a fd_ptr is \p NULL, the file descriptor resulting from the underlying
 * open call is closed after mmap().  Otherwise the file is kept open and the
 * file descriptor is returned in \a fd_ptr.
 *
 * \return Standard.
 *
 * \sa osl_open(), mmap(2).
 */
int mmap_full_file(const char *path, int open_mode, void **map,
		size_t *size, int *fd_ptr)
{
	int fd, ret, mmap_prot, mmap_flags;
	struct stat file_status;

	if (open_mode == O_RDONLY) {
		mmap_prot = PROT_READ;
		mmap_flags = MAP_PRIVATE;
	} else {
		mmap_prot = PROT_READ | PROT_WRITE;
		mmap_flags = MAP_SHARED;
	}
	ret = osl_open(path, open_mode, 0);
	if (ret < 0)
		return ret;
	fd = ret;
	if (fstat(fd, &file_status) < 0) {
		ret = -ERRNO_TO_ERROR(errno);
		goto out;
	}
	*size = file_status.st_size;
	ret = -E_OSL_EMPTY;
	DEBUG_LOG("%s: size %zu\n", path, *size);
	if (!*size)
		goto out;
	*map = mmap(NULL, *size, mmap_prot, mmap_flags, fd, 0);
	if (*map == MAP_FAILED) {
		*map = NULL;
		ret = -E_OSL_MMAP;
		goto out;
	}
	ret = 1;
out:
	if (ret < 0 || !fd_ptr)
		close(fd);
	else
		*fd_ptr = fd;
	return ret;
}

/**
 * A wrapper for munmap(2).
 *
 * \param start The start address of the memory mapping.
 * \param length The size of the mapping.
 *
 * \return Positive on success, \p -E_MUNMAP on errors.
 *
 * \sa munmap(2), mmap_full_file().
 */
int para_munmap(void *start, size_t length)
{
	int err;
	if (munmap(start, length) >= 0)
		return 1;
	err = errno;
	ERROR_LOG("munmap (%p/%zu) failed: %s\n", start, length,
		strerror(err));
	return -ERRNO_TO_ERROR(err);
}
